package ca.uwaterloo.cs.gmysql;

import java.util.zip.ZipInputStream;
import java.util.zip.GZIPInputStream;
import java.util.jar.JarInputStream;
import java.util.zip.ZipEntry;
import java.util.Calendar;
import java.io.*;

import ca.uwaterloo.cs.gmysql.Version;
import ca.uwaterloo.cs.gmysql.StringCache;
import ca.uwaterloo.cs.gmysql.data.*;

public class TAFileReader {

	// Stream tokenizer stuff
	
	private static final byte	CT_DELIMIT	   = 2;
	private static final byte	CT_QUOTE       = 3;
	public  static final int	TT_EOF         = -1;				// The End-of-file token.
	public	static final int	TT_WORD        = -2;				// The word token.	This value is in sval.

	private static char			m_ctype[];

	private boolean				m_drop    = false;
	private boolean				m_lists   = false;
	private boolean				m_tiny_table = false;
	private String				m_input   = null;
	private String              m_schema  = null;
	private String				m_version = null;

	private LineNumberReader	m_reader;
	private String				m_filename;
	private int					m_peekc = ' ';	// Will appear to be first character but ignored so ok
	private boolean				m_pushedBack;	// True if last token pushed back to be reconsumed on next call
	private boolean				m_escaped;


	/* TODO: Thinks about a more memory efficient -- extensible method perhaps by using StringBuffer 
	         Also try to make this static by changing how include is implemented
	 */

	private StringBuffer	m_buf  = new StringBuffer(10240);		// The area that captures a token or attribute value

	/**
	 * The type of the last token returned when m_pushBack is true.	 It's value will either
	 * be one of the following TT_* constants, or a single character.  For example, if '+' is
	 * encountered and is not a valid word character, ttype will be '+'
	 */

	private int 				m_ttype = ' ';
	private char				m_quote = 0;


	/**
	 * The Stream value.
	 */

	private String m_sval;

	private final String filename()
	{
		return m_filename;
	}

	/** Return the current line number. */

	private final int lineno() 
	{
		return m_reader.getLineNumber();
	}

	/* We presume that the reader is sensibly buffered externally */

	private final int charToken() throws IOException 
	{
		int	c;

		c         = m_peekc;
		m_peekc   = m_reader.read();
		switch (c) {
		case '\\':
			switch (m_peekc) {
			case 'n':
				c = '\n';
				break;
			case 't':
				c = '\t';
				break;
			case 'f':
				c = '\f';
				break;
			case 'r':
				c = '\r';
				break;
			case 'e':
				c = 27;
				break;
			case 'd':
				c = 127;
			case '\\':
				m_peekc = ' ';
				break;
			case '"':
				c = '"';
				m_escaped = true;
				break;
			case '\'':
				c = '\'';
				m_escaped = true;
				break;
			default:
				c = m_peekc;
			}
			m_peekc = m_reader.read(); // Consume and discard m_peekc
			break;
		case '\'':
		case '"':
			m_escaped = false;
		}
//		System.err.print("" + ((char) c));
		return c;
	}

	/** Does an unget of the last token */

	protected final void pushBack(int ttype) 
	{
		m_ttype      = ttype;
		m_pushedBack = true;
	} 

	/**
	 * Parses a token from the input stream.  The return value is
	 * the same as the value of ttype.	Typical clients of this
	 * class first set up the syntax tables and then sit in a loop
	 * calling nextToken to parse successive tokens until TT_EOF
	 * is returned.
	 */

	public final int nextToken() throws IOException 
	{
		if (m_pushedBack) {
			m_pushedBack = false;
			return m_ttype;
		}

		int		pos = 0;
		int		c, quote;
		
		for (m_sval = null; ; ) {
			// This loop is repeated while initial whitespace
			c = charToken();

			switch(c) {
			case TT_EOF:
				return TT_EOF;
			case '\f':
			case '\r':
			case '\n':
			case '\t':
			case ' ':
				continue;
			case '"':
			case '\'':
				quote   = c;
				m_quote = (char) quote;
				m_buf.setLength(0);
				for (;;) {
					c = charToken();
					if (c == TT_EOF) {
						System.err.println("Unterminated " + quote + " in input");
						return c;
					}
					if (c == quote && !m_escaped) {
						break;
					}
					m_buf.append((char) c);
				}
				m_sval = m_buf.toString();
				return TT_WORD;
			case '/':
				switch (m_peekc) {
					case '/':
					{
						// Toss the remainder of the line
						for (;;) {
							c = charToken();
							switch (c) {
							case '\r':
								if (m_peekc == '\n') {
									continue;
								}
							case '\n':
							case TT_EOF:
								break;
							default:
								continue;
							}
							break;
						}
						continue;
					}
					case '*':
					{
						for (;;) {
							c = charToken();
							switch (c) {
							case '*':
								if (m_peekc != '/') {
									continue;
								}
								c = charToken();
							case TT_EOF:
								break;
							default:
								continue;
							}
							break;
						}
						continue;
				}	}
			default:
				if (c < m_ctype.length && m_ctype[c] == CT_DELIMIT) {
					m_sval = String.valueOf((char) c);
					return c;
				}

				// Read characters until delimitter

				m_buf.setLength(0);
				m_quote = 0;
				for (;;) {
					m_buf.append((char) c);
					if (m_peekc == TT_EOF) {
						break;
					}
					if (c == '\\') {
						m_buf.append((char) charToken());
					} 
					c = m_peekc;
					if (c < m_ctype.length && m_ctype[c] != 0) {
						break;
					}
					c = charToken();
				}
				m_sval = m_buf.toString();
				if (m_sval.equals(":")) {
					return ':';
				}
				return TT_WORD;
			}
		}
	}

	public final static int SCHEME_TUPLE		= 0;
	public final static int SCHEME_ATTRIBUTE	= 1;
	public final static int FACT_TUPLE			= 2;
	public final static int FACT_ATTRIBUTE		= 3;

	public final static int ERROR				= -1;
	public final static int NONE				= -2; 
	public final static int EOF					= 99;	
	
	public final static String INSTANCE_ID				= "$INSTANCE";	
	public final static String INHERIT_RELN				= "$INHERIT";
	public final static String CONTAIN_ID				= "contain"; 
	public final static String ROOT_ID                  = "$ROOT";
	public final static String SCHEME_ID                = "SCHEME";
	public final static String FACT_ID	                = "FACT";
	public final static String TUPLE_ID	                = "TUPLE";
	public final static String ATTRIBUTE_ID             = "ATTRIBUTE";

	public static boolean		m_strict_TA			    = false;		// Set by -s option

	private   Object			m_context;
	private	  String			m_zipEntry;

	protected String			m_resString         = null;

	protected boolean m_fatalError = false;

	// push backed tokens 

	protected String m_token1, m_token2; 
	protected int	 m_ttype3;
	
	// triple values
	
	public String	m_verb;
	public String	m_object;
	public String	m_subject; 
	public int		m_relations;
	
	// Start line no
	
	protected int	m_startLineno = -1;

	protected void skipRecord() throws IOException 
	{
		int ttype;

		do {
			ttype = nextToken();
		} while (ttype != TT_EOF && ttype != '}');
	}


	// --------------

	// Public methods

	// --------------

	public int getStartLineno()
	{
		return m_startLineno;
	}
	
	public void errorNS(String msg) 
	{
		System.err.println("*** Error (" + filename() + ":" + lineno() + "): " + msg);
	}

	public void error(String msg) 
	{
		System.err.println("*** Error (" + filename() + ":" + lineno() + "): " + msg + ". Found " + m_sval);
	}

	public void warning(String msg) 
	{
		System.err.println(">>> Warning (" + filename() + ":" + lineno() + "): " + msg);
	}

	public void sqlAttribute(boolean schema, boolean relation, String token1, String token2, String token3, String attributeName, String value)
	{
		if (schema) {
			if (relation) {
				new Schema_edge_attributes(token1, attributeName, value);
			} else {
				new Schema_node_attributes(token1, attributeName, value);
			}
		} else {
			if (relation) {
				new Edge_attributes(token1, token2, token3, attributeName, value);
			} else {
				new Node_attributes(token1, attributeName, value);
			}
		}
	}

	public void sqlListAttribute(boolean schema, boolean relation, String token1, String token2, String token3, String attributeName, int pos, String value)
	{
		if (schema) {
			if (relation) {
				new Schema_edge_list_attributes(token1, attributeName, pos, value);
			} else {
				new Schema_node_list_attributes(token1, attributeName, pos, value);
			}
		} else {
			if (relation) {
				new Edge_list_attributes(token1, token2, token3, attributeName, pos, value);
			} else {
				new Node_list_attributes(token1, attributeName, pos, value);
			}
		}
	}


	public int nextSection() throws IOException 
	{
		// Called when a section id is expected
		// Returns section id if new section found, or 
		// EOF when at stream end.

		int ttype;

		String	graph;
		String	type;

		if (m_fatalError) {
			return EOF;
		}

		if (m_token1 != null) {
			if (m_ttype3 == ':') {
				graph  = m_token1; 
				type   = m_token2;
				m_token1 = null;
				m_token2 = null;
			} else {
				error("Expecting section header ':'"); 
				return ERROR; 
			}
		} else { 
			ttype = nextToken();
			switch (ttype) {
			case TT_EOF:
				return EOF;
			case TT_WORD:
				graph = m_sval;

				if (nextToken() != TT_WORD) { 
					error("Expecting section type id");
					return ERROR;
				}
				type = m_sval; 
				if (nextToken() != ':') {
					error("Expecting ':'");
					return ERROR;
				}
				break;
			default:
				error("Expecting graph id");
				return ERROR;
		}	} 

		if (graph.equals(SCHEME_ID)) {
			if (type.equals(TUPLE_ID)) {
				return SCHEME_TUPLE;
			}
			if (type.equals(ATTRIBUTE_ID)) {
				return SCHEME_ATTRIBUTE;
			}
			error("Bad section type");
			return ERROR; 
		}

		if (graph.equals(FACT_ID)) {
			if (type.equals(TUPLE_ID)) {
				return FACT_TUPLE;
			}
			if (type.equals(ATTRIBUTE_ID)) {
				return FACT_ATTRIBUTE;
			}
			error("Bad section type");
			return ERROR;
		}

		error("Bad section id");
		return ERROR;
	}

	public boolean nextSchemaTriple() throws IOException 
	{
		if (!m_fatalError) {

			m_startLineno = lineno();
			
			switch (nextToken()) {
			case TT_EOF:
				// End of section 
				return false;
			case TT_WORD: 
				break;
			default:
				error("Expecting word verb token");
				m_fatalError = true;
				return false;
			}

			m_relations = 0;
			m_verb      = m_sval;

			switch (nextToken()) {
			case TT_WORD:
				m_object = m_sval;
				break;
			case '(':
				m_relations |= 1;
				if (nextToken() == TT_WORD) {
					m_object = m_sval;
					if (nextToken() == ')') {
						break;
				}	}
			default:
				error("Expecting word object token"); 
				m_fatalError = true;
				return false;
			}

			switch (nextToken()) {
			case ':':
				m_token1 = m_verb;
				m_token2 = m_object; 
				m_ttype3 = ':'; 
				return false; 
			case TT_WORD:
				m_subject = m_sval;
				return true;
			case '(':
				m_relations |= 2;
				if (nextToken() == TT_WORD) {
					m_subject = m_sval;
					if (nextToken() == ')') {
						return true;
				}	}
			default:
				error("Expecting word subject token");
				m_fatalError = true;
			}
			return true;
		}
		return false;
	}

	public boolean nextFactTriple() throws IOException 
	{
		if (!m_fatalError) {

			m_startLineno = lineno();

			int ttype = nextToken();

			switch (ttype) {
			case TT_EOF:
				// End of section 
				return false;
			case TT_WORD: 
				m_verb = m_sval;

				if (nextToken() != TT_WORD) {
					error("Expecting word token"); 
					break;
				}

				m_object = m_sval;

				ttype = nextToken();

				if (ttype == ':') {
					m_token1 = m_verb;
					m_token2 = m_object; 
					m_ttype3 = ttype; 
					return false; 
				} 

				if (ttype != TT_WORD) {
					error("Error in tuple parse");
					break;
				}

				m_subject = m_sval;

				return true;
			default:		
				error("Expecting word token");
			}
			m_fatalError = true;
		}
		return false;
	}

	// ------------------
	// TA Reading methods
	// ------------------

	public final void nextAVI(boolean schema, boolean relation, String token1, String token2, String token3, String attributeName) throws IOException 
	{
		int					ttype;
		String				string;

		ttype = nextToken();
		switch (ttype) {
		case '(':
		{
			StringBuffer 	buf       = new StringBuffer(10240);
			int				depth     = 1;
			char			separator = 0;
			int				pos       = 0;
			int				i;
			char			c;

			buf.append('(');
			for (;;) {
				ttype = nextToken();
/*
				System.err.print("Token ");
				switch (m_quote) {
				case '"':
					System.err.print("dquote ");
					break;
				case '\'':
					System.err.print("quote ");
				}

				switch (ttype) {
				case TT_WORD:
					System.err.print("WORD " + m_sval);
					break;
				case TT_EOF:
					System.err.print("EOF");
					break;
				default:
					System.err.print("'" + (char) ttype + "' (" + ttype + ")");
				}
				System.err.println("");
*/

				switch(ttype) {
				case TT_WORD:
				case ':':
					if (m_quote == 0 && separator != 0 && m_sval.equals(",")) {
						// Cludge to fix bug in TA (Comma separated list rather than space separated list)
						continue;
					}
					++pos;
					if (m_lists) {
						sqlListAttribute(schema, relation, token1, token2, token3, attributeName, pos, m_sval);
					}

					if (separator != 0) {
						buf.append(separator);
					}
					if (m_quote != 0) {
						buf.append(m_quote);
						for (i = 0; i < m_sval.length(); ++i) {
							c = m_sval.charAt(i);
							if (c == m_quote) {
								buf.append('\\');
							}
							buf.append(c);
						}
						buf.append(m_quote);
					} else {
						buf.append(m_sval);
					}
					separator = ' ';
					continue;
				case ')':
					buf.append(')');
					break;
				case TT_EOF:
					System.err.println("Unexpected EOF where attribute list value expected at " + lineno());
					return;
				default:
					System.err.println("Unexpected " + ((char) ttype) + " where attribute list value expected at " + lineno());
					return;
				}
				break;
			}
			string = buf.toString();
			// System.err.println(string);
			break;
		}
		case TT_WORD:
		case ':':
			string = m_sval;
			break;
		case TT_EOF:
			System.err.println("Unexpected EOF where attribute value expected at " + lineno());
			return;
		default:
			System.err.println("Unexpected " + ((char) ttype) + " where attribute value expected at " + lineno());
			return;
		}
		sqlAttribute(schema, relation, token1, token2, token3, attributeName, string);
	}

	private void parseAttributes(boolean schema, boolean relation, String token1, String token2, String token3) throws IOException 
	{
		int						ttype;
		String					attributeName;

		for (;;) {
			ttype = nextToken();

			switch (ttype) {
			case '}':
				return;
			case TT_WORD:
				attributeName = m_sval;
				ttype = nextToken();
				if (ttype != '=') {
					// Attribute declaration with no value 
					pushBack(ttype);
					sqlAttribute(schema, relation, token1, token2, token3, attributeName, null);
				} else {
					nextAVI(schema, relation, token1, token2, token3, attributeName);
				}
				break;
			default:
				error("Expecting attribute id");
				return;
			}
		}
	}

	private void 
	processSchemeTuples() throws IOException 
	{
		String			verb, object, subject;
		String			msg;

		while (nextSchemaTriple()) {

			verb    = m_verb;
			object  = m_object;
			subject = m_subject;

			if (verb.equals(INHERIT_RELN)) {							// $INHERIT
				switch (m_relations) {
				case 0:	/* Neither term is a relation */
					if (object.equals("$ENTITY")) {	// $ENTITY
						errorNS("Improper use of $ENTITY with $INHERIT");
						break;
					}
					new Schema_node_inherits(object, subject);
					break;
				case 1:	/* First  is relation		*/
				case 2:	/* Second is relation		*/
					errorNS("Mismatched entity/relation with $INHERIT");
					break;
				case 3:	/* Both term is a relation	*/
					if (object.equals("$RELATION")) {	// $ENTITY
						errorNS("Improper use of $RELATION with $INHERIT");
						break;
					}
					new Schema_edge_inherits(object, subject);
					break;
				}

			} else { 
				if (m_relations != 0) {
					errorNS("Cannot create relationships between relationships");
				} else {
					new Schema_edge_connects(verb, object, subject);
				}
			}
		}
	}

	private void processSchemeAttributes() throws IOException 
	{

		int				ttype;
		String			token1;
		String			token2;
		String			token3;
		String			msg;

		// 
		//  * id "{" {attribute} "}" *
		// 

		while (!m_fatalError) {
		
			m_startLineno = lineno();

			ttype = nextToken();
			msg   = null;

			switch (ttype) {
			case TT_EOF:
				// End of section
				return; 
			case '(':
				// Relation or relation class 
				ttype = nextToken();
				if (ttype != TT_WORD) {
					msg = "Expected ( <relation class name>";
					break;
				} 
				token1 = m_sval;
				ttype  = nextToken();
				if (ttype != ')') {
					msg = "Expected ( " + token1 + "<)>";
					break;
				}
				ttype = nextToken();
				if (ttype != '{') {
					msg = "Expected (...) <{>";
					break;
				}
				parseAttributes(true, true, token1, null, null); 
				break;
			case TT_WORD:
				// Entity class or entity 
				token1 = m_sval;
				ttype  = nextToken();
				if (ttype != '{') {
					if (ttype != TT_WORD) {
						msg = "Expecting " + token1 + " <{>";
						break;
					}
					token2 = m_sval;
					ttype = nextToken();
					if (ttype != ':') {
						msg = "Expecting section header or id <{>";
						break;
					}
					m_token1 = token1; 
					m_token2 = token2;
					m_ttype3 = ttype;
					// End of section 
					return;
				}
				
				parseAttributes(true, false, token1, null, null); 
				break;
			default:
				msg = "Expecting object id";
			}
			if (msg != null) {
				errorNS(msg);
				skipRecord();
			} 
	}	}
	
	private void processFactTuples() throws IOException 
	{
		String		subject;

		while (nextFactTriple()) {
			if (m_verb.equals("$INSTANCE")) {
				new Nodes(m_object, m_subject);
			} else {
				new Edges(m_verb, m_object, m_subject);
			}
		}
	}

	public void processFactAttributes() throws IOException 
	{
		int				ttype;
		String			token1;
		String			token2;
		String			token3;
		String			msg;
		boolean			relation;

		// 
		//  * id "{" {attribute} "}" *
		// 

		while (!m_fatalError) {

			m_startLineno = lineno();

			ttype = nextToken();
			msg   = null;

			switch (ttype) {
			case TT_EOF:
				// End of section
				return; 
			case '(':
			{
				// Relation or relation class 
				relation = true;
				ttype = nextToken();
				if (ttype != TT_WORD) {
					msg = "Expected ( <class> src dst )";
					break;
				} 
				token1 = m_sval;
				ttype = nextToken();
				if (ttype != TT_WORD) {
					msg = "Expected (" + token1 + " <src> dst)" + " not " + ttype;
					break;
				}
				token2 = m_sval;
				ttype = nextToken();
				if (ttype != TT_WORD) {
					msg = "Expected (" + token1 + " " + token2 + " <dst>)" + " not " + ttype;
					break;
				}
				token3 = m_sval;
				ttype = nextToken();
				if (ttype != ')') {
					msg = "Expected (" + token1 + " " + token2 + " " + token3 + "<)>" + " not " + ttype;
					break;
				}
				
				ttype = nextToken();
				if (ttype != '{') {
					msg = "Expected (...) <{>" + " not " + ttype;
					break;
				}
				parseAttributes(false, true, token1, token2, token3); 
				break;
			}
			case TT_WORD:
				// Entity class or entity 
				relation = false;
				token1 = m_sval;
				ttype  = nextToken();
				if (ttype != '{') {
					if (ttype != TT_WORD) {
						msg = "Expecting " + token1 + " <{>";
						break;
					}
					token2 = m_sval;
					ttype = nextToken();
					if (ttype != ':') {
						msg = "Expecting section header or id <{>";
						break;
					}
					m_token1 = token1; 
					m_token2 = token2;
					m_ttype3 = ttype;
					// End of section 
					return;
				}
				parseAttributes(false, false, token1, null, null);
				break;
			default:
				msg = "Expecting object id";
			}
			if (msg != null) {
				errorNS(msg);
				skipRecord();
			} 
	}	} 

	private void parseStream(LineNumberReader is) 
	{
		try {
			for (;;) {
				switch(nextSection()) {
				case SCHEME_TUPLE:
					processSchemeTuples();
					continue;
				case SCHEME_ATTRIBUTE:
					processSchemeAttributes();
					continue;
				case FACT_TUPLE:
					processFactTuples();
					continue;
				case FACT_ATTRIBUTE:
					processFactAttributes();
					continue;
				case EOF:
					return;
				}
			}

		} catch (Exception e) {
			m_resString = e.getMessage();
			System.err.println("Parse error      : " + m_resString);
			System.err.println("Start Line number: " + getStartLineno());
			System.err.println("Last line read   : " + lineno());
//			e.printStackTrace();
		}	
	}

	private InputStream decompress(InputStream is, String source, String subfile)
	{
		InputStream ret      = is;
		int			lth      = source.length();

		if (lth > 4) {
			String ends = source.substring(lth - 4);
			if (ends.equalsIgnoreCase(".zip")) {
				ZipInputStream	zipInputStream;
				ZipEntry		zipEntry;

				try {
					zipInputStream = new ZipInputStream(is);
					for (;;) {
						zipEntry = zipInputStream.getNextEntry();
						if (subfile == null || subfile.equalsIgnoreCase(zipEntry.getName())) {
							break;
						}
						zipInputStream.closeEntry();
					}
				} catch (Exception e) {
					System.err.println("Attempt to open " + source + ((subfile == null) ? "" : "#" + subfile) + " as zip file failed");
					m_resString    = e.getMessage();
					zipInputStream = null;
				}
				return zipInputStream;
			}
			if (ends.equalsIgnoreCase(".jar")) {
				JarInputStream	jarInputStream;
				ZipEntry		zipEntry;

				try {
					jarInputStream = new JarInputStream(is);
					for (;;) {
						zipEntry = jarInputStream.getNextEntry();
						if (subfile == null || subfile.equalsIgnoreCase(zipEntry.getName())) {
							break;
						}
						jarInputStream.closeEntry();
					}
				} catch (Exception e) {
					System.err.println("Attempt to open " + source + ((subfile == null) ? "" : "#" + subfile) + " as jar file failed");
					m_resString    = e.getMessage();
					jarInputStream = null;
				}
				return jarInputStream;
			}

			if (lth > 5) {
				ends = source.substring(lth - 5);
				if (ends.equalsIgnoreCase(".gzip")) {
					GZIPInputStream	gzipInputStream;

					try {
						gzipInputStream = new GZIPInputStream(is);
					} catch (Exception e) {
						System.err.println("Attempt to open " + source + " as gzip file failed");
						m_resString     = e.getMessage();
						gzipInputStream = null;
					}
					return gzipInputStream;
		}	}	}
		return is;
	}

	protected void parseFile(String taPath) 
	{
		String	entry = null;
		int		lth   = taPath.length();
		File	file  = null;

		if (lth > 0) {
			char	lc	  = taPath.charAt(lth-1);

			if (lth > 2 &&  lc == ']') {
				int	i = taPath.lastIndexOf('['); 

				if (i > 0 && i < lth - 2) {
					entry  = taPath.substring(i+1, lth-1);
					if (m_zipEntry == null) {
						m_zipEntry = entry;
					}
					taPath = taPath.substring(0, i);
					lth    = i;
					lc     = taPath.charAt(lth-1);
			}	}
			if (lc == File.separatorChar) {
				taPath = taPath.substring(0, lth-1);
		}	}	
		
		try {
			InputStream	is   = null;

			if (lth > 0) {
				file = new File(taPath);
				if (file != null) {
					is = new FileInputStream(file);
					is = decompress(is, taPath, entry);
			}	}
			if (is == null) {
				if (m_resString == null) {
					m_resString = "No input stream specified";
				}
				return;
			} 
			InputStreamReader	reader       = new InputStreamReader(is);
			LineNumberReader	linenoReader = new LineNumberReader(reader);
				
			m_reader = linenoReader;

			linenoReader.setLineNumber(1);

			parseStream(linenoReader);
			linenoReader.close();
			linenoReader = null;
			
		} catch (Exception e) {
			m_resString = e.getMessage();
		}
	}

    public TAFileReader() 
    {
		m_ctype = new char[256];
		m_ctype[' ']  = CT_DELIMIT;
		m_ctype['\f'] = CT_DELIMIT;
		m_ctype['\t'] = CT_DELIMIT;
		m_ctype['\r'] = CT_DELIMIT;
		m_ctype['\n'] = CT_DELIMIT;
		m_ctype['=']  = CT_DELIMIT;
//		m_ctype[':']  = CT_DELIMIT;
		m_ctype['{']  = CT_DELIMIT;
		m_ctype['}']  = CT_DELIMIT;
		m_ctype['(']  = CT_DELIMIT;
		m_ctype[')']  = CT_DELIMIT;
		m_ctype['"']  = CT_QUOTE;
		m_ctype['\''] = CT_QUOTE;
    }
    
    // Returning null indicates success
    
    public void read(String taFile)
    {
		m_resString = null;
		m_zipEntry  = null;
		m_filename  = taFile;
		
		parseFile(taFile);
		if (m_resString != null) {
			System.err.println(m_resString);
		}
		return;
	}

	public void drop_tables()
	{
		String[] tables = {
			"schema_node_inherits",
			"schema_edge_inherits",
			"schema_edge_connects",
			"schema_node_attributes",
			"schema_edge_attributes",
			"nodes",
			"edges",
			"node_attributes",
			"edge_attributes",
			"schema_node_list_attributes",
			"schema_edge_list_attributes",
			"node_list_attributes",
			"edge_list_attributes",
			"ids"
		};

		int i;

		for (i = 0; i < tables.length; ++i) {
			System.out.print("drop table if exists ");
			if (m_schema != null) {
				System.out.print(m_schema);
			}
			System.out.print(tables[i]);
			System.out.println(";");
		}
		System.out.println("");
	}

	public void create_tables()
	{
		String schema  = m_schema;
		String version = m_version;

		Schema_node_inherits.create(schema, version);
		Schema_edge_inherits.create(schema, version);
		Schema_edge_connects.create(schema, version);
		Schema_node_attributes.create(schema, version);
		Schema_edge_attributes.create(schema, version);
		Nodes.create(schema, version);
		Edges.create(schema, version);
		Node_attributes.create(schema, version);
		Edge_attributes.create(schema, version);


		if (m_lists) {
			Schema_node_list_attributes.create(schema, version);
			Schema_edge_list_attributes.create(schema, version);
			Node_list_attributes.create(schema, version);
			Edge_list_attributes.create(schema, version);
		}
		if (m_tiny_table) {
			Ids.create(schema, version);
		}
	}

	public void dump_tables()
	{
		String schema  = m_schema;
		String version = m_version;

		Schema_node_inherits.dump(schema, version);
		Schema_edge_inherits.dump(schema, version);
		Schema_edge_connects.dump(schema, version);
		Schema_node_attributes.dump(schema, version);
		Schema_edge_attributes.dump(schema, version);
		Nodes.dump(schema, version);
		Edges.dump(schema, version);
		Node_attributes.dump(schema, version);
		Edge_attributes.dump(schema, version);

		if (m_lists) {
			Schema_node_list_attributes.dump(schema, version);
			Schema_edge_list_attributes.dump(schema, version);
			Node_list_attributes.dump(schema, version);
			Edge_list_attributes.dump(schema, version);
		}
		if (m_tiny_table) {
			Ids.dump(schema, version);
		}
	}
	
	public void augment_tables()
	{
		String schema  = m_schema;
		String version = m_version;

		Schema_node_inherits.augment(schema, version);
		Schema_edge_inherits.augment(schema, version);
		Schema_edge_connects.augment(schema, version);
		Schema_node_attributes.augment(schema, version);
		Schema_edge_attributes.augment(schema, version);
		Nodes.augment(schema, version);
		Edges.augment(schema, version);
		Node_attributes.augment(schema, version);
		Edge_attributes.augment(schema, version);

		if (m_lists) {
			Schema_node_list_attributes.augment(schema, version);
			Schema_edge_list_attributes.augment(schema, version);
			Node_list_attributes.augment(schema, version);
			Edge_list_attributes.augment(schema, version);
		}
		if (m_tiny_table) {
			Ids.augment(schema, version);
		}
	}

	public void index_tables()
	{
		String schema  = m_schema;
		String version = m_version;

		Schema_node_inherits.index(schema, version);
		Schema_edge_inherits.index(schema, version);
		Schema_edge_connects.index(schema, version);
		Schema_node_attributes.index(schema, version);
		Schema_edge_attributes.index(schema, version);
		Nodes.index(schema, version);
		Edges.index(schema, version);
		Node_attributes.index(schema, version);
		Edge_attributes.index(schema, version);

		if (m_lists) {
			Schema_node_list_attributes.index(schema, version);
			Schema_edge_list_attributes.index(schema, version);
			Node_list_attributes.index(schema, version);
			Edge_list_attributes.index(schema, version);
		}
		if (m_tiny_table) {
			Ids.index(schema, version);
		}
	}

	// http://guilhembichot.blogspot.ca/2013/11/with-recursive-and-mysql.html
	// http://stackoverflow.com/questions/16513418/how-to-do-the-recursive-select-query-in-mysql
	
	public void create_compute_paths()
	{
		String schema = m_schema;

		System.out.print  ("DROP PROCEDURE IF EXISTS ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_paths;");
		System.out.println("");
		System.out.println("delimiter %");
		System.out.println("");
		System.out.print  ("CREATE PROCEDURE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_paths(");
		System.out.println("  seed   varchar(255),");
		System.out.println("  target varchar(255),");
		System.out.println("  rule   varchar(10000)");
		System.out.println(")");
		System.out.println("BEGIN");
		System.out.println("");
		System.out.println("  DECLARE new_rows int;");
		System.out.println("");
		System.out.println("  SET @str = CONCAT('DROP TABLE IF EXISTS ', target);");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.println("  SET @str = CONCAT('CREATE TABLE ', target, '(start varchar(255) not null, end varchar(255) not null, unique index btree (start,end)) default character set utf8 ENGINE=Innodb');");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.println("  SET @str = CONCAT('INSERT IGNORE INTO ', target, ' select start, end from ',seed);");
		System.out.println("  recursion : LOOP");
		System.out.println("    PREPARE stmt FROM @str;");
		System.out.println("    EXECUTE stmt; ");
		System.out.println("    select row_count() into new_rows;");
// "    select concat(new_rows, ' new rows found');");
		System.out.println("    if not new_rows then");
		System.out.println("      leave recursion;");
		System.out.println("    end if;");
		System.out.println("    SET @str = CONCAT('insert ignore into ', target, ' select start, end from (select t1.start, t2.end from ', target, ' t1 join ', target, ' t2 where t1.end = t2.start) t3 ', rule);");
		System.out.println("");
		System.out.println("  END LOOP;");
		System.out.println("END %");
		System.out.println("");
	}

	public void create_compute_pathsteps()
	{
		String schema = m_schema;

		System.out.println("delimiter ;");
		System.out.println("");
		System.out.print  ("DROP PROCEDURE IF EXISTS ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_pathsteps;");
		System.out.println("");
		System.out.println("delimiter %");
		System.out.println("");
		System.out.print  ("CREATE PROCEDURE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_pathsteps(");
		System.out.println("  seed   varchar(255),");
		System.out.println("  target varchar(255),");
		System.out.println("  rule   varchar(10000)");
		System.out.println(")");
		System.out.println("BEGIN");
		System.out.println("");
		System.out.println("  DECLARE new_rows int;");
		System.out.println("");
		System.out.println("  SET @str = CONCAT('DROP TABLE IF EXISTS ', target);");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.println("  SET @str = CONCAT('CREATE TABLE ', target, '(steps int unsigned not null, start varchar(255) not null, end varchar(255) not null, unique index btree (start,end)) default character set utf8 ENGINE=Innodb');");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.println("  SET @str = CONCAT('INSERT IGNORE INTO ', target, ' select 1, start, end from ',seed);");
		System.out.println("  recursion : LOOP");
		System.out.println("    PREPARE stmt FROM @str;");
		System.out.println("    EXECUTE stmt; ");
		System.out.println("    select row_count() into new_rows;");
// "    select concat(new_rows, ' new rows found');");
		System.out.println("    if not new_rows then");
		System.out.println("      leave recursion;");
		System.out.println("    end if;");
		System.out.println("    SET @str = CONCAT('insert ignore into ', target, ' select steps, start, end from (select t1.steps+t2.steps as steps, t1.start, t2.end from ', target, ' t1 join ', target, ' t2 where t1.end = t2.start order by steps) t3 ',rule);");
		System.out.println("");
		System.out.println("  END LOOP;");
		System.out.println("END %");
		System.out.println("");
		System.out.println("delimiter ;");
		System.out.println("");
	}
	
	public void create_compute_schema_all_node_attributes()
	{
		String schema = m_schema;

		System.out.print  ("DROP PROCEDURE IF EXISTS ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_schema_all_node_attributes;");
		System.out.println("");
		System.out.println("delimiter %");
		System.out.println("");
		System.out.print  ("CREATE PROCEDURE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_schema_all_node_attributes()");
		System.out.println("BEGIN");
		System.out.println("");
		System.out.println("  DECLARE new_rows int;");
		System.out.println("");
		System.out.print  ("  SET @str = CONCAT('DROP TABLE IF EXISTS ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_all_node_attributes');");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.print  ("  SET @str = CONCAT('CREATE TABLE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_all_node_attributes (class varchar(255) not null, superclass varchar(255), attribute varchar(255) not null, value varchar(10000), unique index (class,attribute)) default character set utf8 ENGINE=Innodb');");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.print  ("  SET @str = CONCAT('INSERT IGNORE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.print  ("schema_all_node_attributes select class,null,attribute,value  from ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_node_attributes');");
		System.out.println("  recursion : LOOP");
		System.out.println("    PREPARE stmt FROM @str;");
		System.out.println("    EXECUTE stmt; ");
		System.out.println("    select row_count() into new_rows;");
		// System.out.println("    select concat(new_rows, ' new rows found');");
		System.out.println("    if not new_rows then");
		System.out.println("      leave recursion;");
		System.out.println("    end if;");
		System.out.print  ("    SET @str = CONCAT('insert ignore into ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.print  ("schema_all_node_attributes select t2.class, t2.superclass, t1.attribute, t1.value from ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.print  ("schema_all_node_attributes t1 join ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_node_inherits t2 where t1.class = t2.superclass');");
		System.out.println("");
		System.out.println("  END LOOP;");
		System.out.println("END %");
		System.out.println("");
		System.out.println("delimiter ;");
		System.out.println("");
	}

	public void create_compute_schema_all_edge_attributes()
	{
		String schema = m_schema;

		System.out.print  ("DROP PROCEDURE IF EXISTS ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_schema_all_edge_attributes;");
		System.out.println("");
		System.out.println("delimiter %");
		System.out.println("");
		System.out.print  ("CREATE PROCEDURE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("compute_schema_all_edge_attributes()");
		System.out.println("BEGIN");
		System.out.println("");
		System.out.println("  DECLARE new_rows int;");
		System.out.println("");
		System.out.print  ("  SET @str = CONCAT('DROP TABLE IF EXISTS ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_all_edge_attributes');");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.print  ("  SET @str = CONCAT('CREATE TABLE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_all_edge_attributes (class varchar(255) not null, superclass varchar(255), attribute varchar(255) not null, value varchar(10000), unique index (class,attribute)) default character set utf8 ENGINE=Innodb');");
		System.out.println("  PREPARE stmt FROM @str;");
		System.out.println("  EXECUTE stmt; ");
		System.out.print  ("  SET @str = CONCAT('INSERT IGNORE ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.print  ("schema_all_edge_attributes select class,null,attribute,value  from ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_edge_attributes');");
		System.out.println("  recursion : LOOP");
		System.out.println("    PREPARE stmt FROM @str;");
		System.out.println("    EXECUTE stmt; ");
		System.out.println("    select row_count() into new_rows;");
		// System.out.println("    select concat(new_rows, ' new rows found');");
		System.out.println("    if not new_rows then");
		System.out.println("      leave recursion;");
		System.out.println("    end if;");
		System.out.print  ("    SET @str = CONCAT('insert ignore into ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.print  ("schema_all_edge_attributes select t2.class, t2.superclass, t1.attribute, t1.value from ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.print  ("schema_all_edge_attributes t1 join ");
		if (schema != null) {
			System.out.print(schema);
		}
		System.out.println("schema_edge_inherits t2 where t1.class = t2.superclass');");
		System.out.println("");
		System.out.println("  END LOOP;");
		System.out.println("END %");
		System.out.println("");
		System.out.println("delimiter ;");
		System.out.println("");
	}

	public void create_procedures()
	{
		create_compute_paths();
		create_compute_pathsteps();
		create_compute_schema_all_node_attributes();
		create_compute_schema_all_edge_attributes();
	}

	public void setOptions(String args[]) 
	{
		int		n;
		char	c;
		boolean	ok;
		int		lth;
		String	option, value;

		ok = true;
		if (args == null) {
			System.err.println("No arguments");
			ok = false;
		} else {

			// Process command line switches
	
			for (n = 0; n < args.length; ++n) {
				option = args[n];
				lth    = option.length();
				if (lth < 2 || option.charAt(0) != '-') {
					break;
				}
				c = option.charAt(1);
				switch (c) {
				case 'd':
					m_drop = true;
					continue;
				case 'l':
					m_lists = true;
					continue;
				case 'h':
					Version.Details();
					continue;
				case 't':
					if (Row.g_hash) {
						m_tiny_table = true;
					} 
					Row.g_hash = true;
					continue;
				case '?':
					ok = false;
					break;
				default:
					System.err.println("Unknown option '" + c + "'");
					ok = false;
			}	}
	
			if (ok) {	
				if (n < args.length) {
					m_input = args[n++];
				}
				if (n < args.length) {
					m_schema = args[n++];
				}
				if (n < args.length) {
					m_version = args[n++];
				}
				if (n < args.length) {
					System.err.println("Too many parameters");
					ok = false;
				}
			}
		}
		if (ok) {
			System.out.print("# GMysql ");
			System.out.print(Version.Number());
			System.out.print(" ");
			System.out.print(Version.CompileDate());
			System.out.print(" Started ");
			System.out.println(Calendar.getInstance().getTime());
			System.out.println("");
			if (m_drop) {
				drop_tables();
			}
			read(m_input);
			create_tables();
			dump_tables();
			augment_tables();
			index_tables();
			create_procedures();
			System.out.println("");
			System.out.print("# End of input ");
			System.out.println(Calendar.getInstance().getTime());
			System.exit(0);
		}
		System.err.println("Usage: [-d/rop] [-l/ists] [-t/iny [-t/able]] <input> [<schema/table prefix>] [<version>]");
		System.exit(1);
	}

	public static void main(String[] args) 
	{
		TAFileReader reader = new TAFileReader();

		reader.setOptions(args);
	}
}


